#pragma once

#include <base/heap.h>
#include <base/log.h>
#include <functional>
#include <os/reporter.h>
#include <qumela/types.h>

namespace Topo_mgr {

	using namespace Genode;

	class Generatible;
	
	template <class T>
	class Element : public Fifo<Element<T>>::Element
	{
	private:
		T* _value;

	public:
		T* value() { return _value; }

		Element(T* value)
			: _value(value)
		{
		}
	};
	
	class Generatible
	{
	protected:
		Heap &_heap;

	public:
		virtual void generate(Xml_generator &xml) = 0;

		virtual ~Generatible() { }
		
		Generatible(Heap &heap)
			: _heap(heap)
		{
		}
	};

	template <class T>
	class Generatible_container : public Generatible
	{
	private:
		Fifo<Element<T>> _children { };

	protected:
		void add(T* g)
		{
			_children.enqueue(new (_heap) Element<T>(g));
		}

		bool has_children()
		{
			return !_children.empty();
		}

		void generate_children(Xml_generator &xml)
		{
			for (Element<T>* e = _children.head(); e != nullptr; e = e->next())
			{
				e->value()->generate(xml);
			}
		}

	public:
		virtual ~Generatible_container()
		{
			while (!_children.empty())
			{
				auto c = _children.dequeue();
				destroy(_heap, c->value());
				destroy(_heap, c);
			}
		}

		Generatible_container(Heap &heap)
			: Generatible(heap)
		{
		}
	};	
	
	class Vfs_entry : public Generatible
	{
	public:
		Vfs_entry(Heap &heap)
			: Generatible(heap)
		{
		}
	};
	
	class Vfs_inline : public Vfs_entry
	{
	private:
		Vfs_name _name;
		String<4096> _content;

	public:
		virtual void generate(Xml_generator &xml) override
		{
			xml.node("inline", [&] () { 
				xml.attribute("name", _name);
				xml.append(_content.string());
			});
		}
		
		Vfs_inline(Heap &heap, Vfs_name name, Vfs_content content)
			: Vfs_entry(heap), _name(name), _content(content)
		{
		}
	};
	
	class Vfs_fs : public Vfs_entry
	{
	private:
		Interface_label _label;

	public:
		virtual void generate(Xml_generator &xml) override
		{
			xml.node("fs", [&] () { 
				if (_label != "") {
					xml.attribute("label", _label);
				}
			} );
		}
		
		Vfs_fs(Heap &heap, Interface_label label)
			: Vfs_entry(heap), _label(label)
		{
		}
	};
	
	class Vfs_dir : public Vfs_entry, public Generatible_container<Vfs_entry>
	{
	private:
		Vfs_name _name;

	public:
		Vfs_dir* dir(Vfs_name name)
		{
			Vfs_dir* d = new (Vfs_entry::_heap) Vfs_dir(Vfs_entry::_heap, name);
			add(d);
			return d;
		}

		Vfs_fs* fs(Interface_label label)
		{
			Vfs_fs* f = new (Vfs_entry::_heap) Vfs_fs(Vfs_entry::_heap, label);
			add(f);
			return f;
		}
		
		Vfs_fs* fs()
		{
			return fs("");
		}

		Vfs_inline* file(Vfs_name name, Vfs_content content)
		{
			Vfs_inline* i = new (Vfs_entry::_heap) Vfs_inline(Vfs_entry::_heap, name, content);
			add(i);
			return i;
		}
		
		virtual void generate(Xml_generator &xml) override
		{
			xml.node("dir", [&] () { 
				xml.attribute("name", _name);
				generate_children(xml);
			} );
		}
		
		Vfs_dir(Heap &heap, Vfs_name name)
			: Vfs_entry(heap), Generatible_container(heap), _name(name)
		{
		}
	};
		
	class Service_route : public Generatible
	{
	public:
		Service_route(Heap &heap)
			: Generatible(heap)
		{
		}
	};
	
	class Vfs : public Generatible_container<Vfs_entry>
	{
	public:
		Vfs_dir* dir(Vfs_name name)
		{
			Vfs_dir* d = new (_heap) Vfs_dir(_heap, name);
			add(d);
			return d;
		}

		Vfs_fs* fs(Interface_label label)
		{
			Vfs_fs* f = new (_heap) Vfs_fs(_heap, label);
			add(f);
			return f;
		}
		
		Vfs_fs* fs()
		{
			return fs("");
		}
		
		Vfs_inline* file(Vfs_name name, Vfs_content content)
		{
			Vfs_inline* i = new (_heap) Vfs_inline(_heap, name, content);
			add(i);
			return i;
		}

		virtual void generate(Xml_generator &xml) override
		{
			xml.node("vfs", [&] () { 
				generate_children(xml);
			} );
		}

		Vfs(Heap &heap)
			: Generatible_container(heap)
		{
		}
	};
	
	class Default_policy : public Generatible
	{
	private:
		File_path _path;
		bool _writeable;

	public:
		virtual void generate(Xml_generator &xml) override
		{
			xml.node("default-policy", [&] () { 
				xml.attribute("path", _path);
				xml.attribute("writeable", _writeable ? "yes" : "no");
			} );
		}
		
		Default_policy(Heap &heap, File_path path, bool writeable)
			: Generatible(heap), _path(path), _writeable(writeable)
		{
		}
	};
	
	class Child : public Service_route
	{
	private:
		Component_name _name;
		Interface_label _label;

	public:
		virtual void generate(Xml_generator &xml) override
		{
			xml.node("child", [&] () { 
				xml.attribute("name", _name);
				if (_label != "") {
					xml.attribute("label", _label);
				}
			} );
		}
		
		Child(Heap &heap, Component_name name, Interface_label label)
			: Service_route(heap), _name(name), _label(label)
		{
		}
	};
	
	class Parent : public Service_route
	{
	private:
		Interface_label _label;

	public:
		virtual void generate(Xml_generator &xml) override
		{
			xml.node("parent", [&] () { 
				if (_label != "") {
					xml.attribute("label", _label);
				}
			} );
		}
		
		Parent(Heap &heap, Interface_label label)
			: Service_route(heap), _label(label)
		{
		}
	};
	
	class Service : public Generatible_container<Service_route>
	{
	private:
		Interface_type _name;
		Interface_label _label;
		Interface_label _label_last;

	public:
		virtual void generate(Xml_generator &xml) override
		{
			xml.node("service", [&] () {
				xml.attribute("name", _name);

				if (_label != "") {
					xml.attribute("label", _label);
				}
				
				if (_label_last != "") {
					xml.attribute("label_last", _label_last);
				}

				generate_children(xml);
			} );
		}
		
		Parent* parent()
		{
			Parent* p = new (_heap) Parent(_heap, "");
			add(p);
			return p;
		}
		
		Parent* parent(Interface_label label)
		{
			Parent* p = new (_heap) Parent(_heap, label);
			add(p);
			return p;
		}
		
		Child* child(Component_name name)
		{
			Child* c = new (_heap) Child(_heap, name, "");
			add(c);
			return c;
		}
		
		Child* child(Component_name name, Interface_label label)
		{
			Child* c = new (_heap) Child(_heap, name, label);
			add(c);
			return c;
		}
		
		Service(Heap &heap, Component_name name, Interface_label label, Interface_label label_last)
			: Generatible_container(heap), _name(name), _label(label), _label_last(label_last)
		{
		}
	};
	
	class Start_config : public Generatible_container<Generatible>
	{
	private:
		Constructible<std::function<void(Xml_generator &cfg)>> _construct { };
		Vfs* _vfs = nullptr;

	public:
		Start_config(const Start_config&) = delete;
		Start_config& operator=(const Start_config&) = delete;

		virtual void generate(Xml_generator &xml) override
		{
			xml.node("config", [&] () {
				generate_children(xml);
				if (_construct.constructed()) {
					(*_construct)(xml);
				}
			});
		}

		Vfs* vfs()
		{
			if (_vfs == nullptr) {
				_vfs = new (_heap) Vfs(_heap);
				add(_vfs);
			}

			return _vfs;
		}

		Default_policy* default_policy(File_path path, bool writeable)
		{
			Default_policy* p = new (_heap) Default_policy(_heap, path, writeable);
			add(p);
			return p;
		}

		void build(std::function<void(Xml_generator &cfg)> construct)
		{
			_construct.construct(construct);
		}
		
		Start_config(Heap &heap)
			: Generatible_container(heap)
		{
		}
	};
	
	class Route : public Generatible_container<Service>
	{
	public:
		virtual void generate(Xml_generator &xml) override
		{
			if (has_children()) {
				xml.node("route", [&] () {
					generate_children(xml);
				});
			}
		}
		
		Service* service(Component_name name)
		{
			Service* s = new (_heap) Service(_heap, name, "", "");
			add(s);
			return s;
		}

		Service* service(Component_name name, Interface_label label)
		{
			Service* s = new (_heap) Service(_heap, name, label, "");
			add(s);
			return s;
		}
		
		Service* service(Component_name name, Interface_label label, Interface_label label_last)
		{
			Service* s = new (_heap) Service(_heap, name, label, label_last);
			add(s);
			return s;
		}
		
		Route(Heap &heap)
			: Generatible_container(heap)
		{
		}
	};
	
	class Common_routes : public Generatible_container<Service>
	{
	public:
		virtual void generate(Xml_generator &xml) override
		{
			if (has_children()) {
				xml.node("common_routes", [&] () {
					generate_children(xml);
				});
			}
		}
		
		Service* service(Component_name name)
		{
			Service* s = new (_heap) Service(_heap, name, "", "");
			add(s);
			return s;
		}

		Service* service(Component_name name, Interface_label label)
		{
			Service* s = new (_heap) Service(_heap, name, label, "");
			add(s);
			return s;
		}
		
		Service* service(Component_name name, Interface_label label, Interface_label label_last)
		{
			Service* s = new (_heap) Service(_heap, name, label, label_last);
			add(s);
			return s;
		}
		
		Common_routes(Heap &heap)
			: Generatible_container(heap)
		{
		}
	};

	class Start : public Generatible_container<Generatible>
	{
	private:
		Component_name _name;
		Package_path _package;
		Route* _route = nullptr;
		Start_config* _config = nullptr;

	public:
		Start(const Start&) = delete;
		Start& operator=(const Start&) = delete;

		Start(Heap &heap, Component_name name, Package_path package)
			: Generatible_container(heap), _name(name), _package(package)
		{
			_route = new (_heap) Route(_heap);
			add(_route);
		}

		virtual void generate(Xml_generator &xml) override
		{
				xml.node("start", [&] () {
					xml.attribute("name", _name);
					xml.attribute("pkg", _package);
					generate_children(xml);
				});
		}

		Route* route()
		{
			return _route;
		}
		
		Start_config* config()
		{
			if (_config == nullptr)
			{
				_config = new (_heap) Start_config(_heap);
				add(_config);
			}

			return _config;
		}
	};
	
	class Init_config : public Generatible_container<Generatible>
	{
	private:
		Common_routes* _common_routes = nullptr;
	
	public:
		Init_config(const Init_config&) = delete;
		Init_config& operator=(const Init_config&) = delete;

		Init_config(Heap &heap)
			: Generatible_container(heap)
		{
			_common_routes = new (_heap) Common_routes(_heap);
			add(_common_routes);
		}

		virtual void generate(Xml_generator &xml) override
		{
				xml.attribute("arch", "x86_64");
				generate_children(xml);
		}

		Common_routes* common_routes()
		{
			return _common_routes;
		}

		Start* start(Component_name name, Package_path package)
		{
			Start* s = new (_heap) Start(_heap, name, package);
			add(s);
			return s;
		}
	};
	
}
